# Copyright 2010-2011 Florent Le Coz <louiz@louiz.org>
#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the zlib license. See the COPYING file.

import logging
log = logging.getLogger(__name__)

import os
import curses
import tempfile
import subprocess

import tabs
import windows
import atom_parser

from config import config

from datetime import datetime
from sleekxmpp.xmlstream import ElementBase, ET

ATOM_XMLNS = "http://www.w3.org/2005/Atom"

item_template = \
"""
-- About you
-- Your name
Author: %(author)s

-- Your jid (e.g. xmpp:romeo@example.com) or website’s url (http://example.com)
URI: xmpp:%(jid)s

-- Your email address (e.g. romeo@mail.com)
email: 
Title: 

-- Please use the form dd/mm/yyy hh:mm:ss, or leave empty
-- If you leave the date empty it will automatically be filled with the current time and date (recommended)
Date: 

--- body --- Until the end of file, this will be your item's body

"""

def parse_template(lines):
    """
    takes a template string (splitted by lines) and returns a dict with the correstponding values
    """
    lines = [line.lower() for line in lines]
    res = {}
    reading_body = False
    body = []
    for line in lines:
        if line.startswith('--- body ---'):
            reading_body = True
            continue
        if reading_body:
            body.append(line)
            continue
        if line.startswith('-- '):
            continue
        for value in ('author', 'uri', 'email', 'title', 'date'):
            if line.startswith('%s:' % (value,)):
                res[value] = line.split(':', 1)[1].strip()
    res['body'] = ''.join(body)
    return res

def create_entry_from_dict(dic):
    """
    Takes a dict with the correct values and returns an ET.Element
    representing an Atom entry
    """
    entry_elem = ET.Element('entry', xmlns=ATOM_XMLNS)

    author_elem = ET.Element('author', xmlns=ATOM_XMLNS)
    if dic.get('author'):
        name_elem = ET.Element('name', xmlns=ATOM_XMLNS)
        name_elem.text = dic.get('author')
        author_elem.append(name_elem)
    if dic.get('uri'):
        uri_elem = ET.Element('uri', xmlns=ATOM_XMLNS)
        uri_elem.text = dic.get('uri')
        author_elem.append(uri_elem)
    if dic.get('email'):
        email_elem = ET.Element('email', xmlns=ATOM_XMLNS)
        email_elem.text = dic.get('email')
        author_elem.append(email_elem)
    entry_elem.append(author_elem)
    if dic.get('title'):
        title_elem = ET.Element('title', xmlns=ATOM_XMLNS)
        title_elem.text = dic.get('title')
        entry_elem.append(title_elem)
    # TODO
    # if dic.get('date'):
    date_elem =  ET.Element('published', xmlns=ATOM_XMLNS)
    date_elem.text = '%s' % datetime.now()
    entry_elem.append(date_elem)
    summary_elem = ET.Element('summary', xmlns=ATOM_XMLNS)
    summary_elem.text = dic.get('body') or ''
    entry_elem.append(summary_elem)
    return entry_elem

class PubsubNode(object):
    node_type = None            # unknown yet
    def __init__(self, name, parent=None):
        self.items = []
        self.name = name
        self.parent = parent


class LeafNode(PubsubNode):
    node_type = "leaf"
    def __init__(self, name, parent=None):
        PubsubNode.__init__(self, name, parent)


class CollectionNode(PubsubNode):
    node_type = "collection"
    def __init__(self, name, parent=None):
        PubsubNode.__init__(self, name, parent)
        self.subnodes = []


class PubsubItem(object):
    def __init__(self, idd, content):
        self.id = idd
        self.content = content
        self.parsed_content = atom_parser.parse_atom_entry(content)

    def to_dict(self, columns):
        """
        returns a dict of the values listed in columns
        """
        ret = {}
        for col in columns:
            ret[col] = self.__dict__.get(col) or ''
        if self.parsed_content:
            ret['title'] = self.parsed_content.get('title')
            ret['author'] = self.parsed_content['author'].get('name')
        return ret

class PubsubBrowserTab(tabs.Tab):
    """
    A tab containing a pubsub browser letting the user
    list nodes and items, view, add and delete items, etc
    """
    def __init__(self, server):
        """
        Server is the name of the pubsub server, for example:
        pubsub.example.com
        All action done in this tab will be made on that server.
        """
        tabs.Tab.__init__(self)
        self.current_node = None # the subnode we are listing. None means the root
        self.server = server
        self.nodes = []         # the lower level of nodes

        self.upper_message = windows.Topic()
        self.set_info_message('Loading')

        # Node List View
        node_columns = ('node', 'name',)
        self.node_list_header = windows.ColumnHeaderWin(node_columns)
        self.node_listview = windows.ListWin(node_columns)

        # Item List View
        item_columns = ('title', 'author', 'id')
        self.item_list_header = windows.ColumnHeaderWin(item_columns)
        self.item_listview = windows.ListWin(item_columns)

        # Vertical Separator
        self.vertical_separator = windows.VerticalSeparator()

        # Item viewer
        self.item_viewer = windows.SimpleTextWin('')
        self.default_help_message = windows.HelpText("“c”: create a node.")
        self.input = self.default_help_message

        self.key_func['c'] = self.command_create_node
        self.key_func['p'] = self.command_publish_item
        self.key_func["M-KEY_DOWN"] = self.scroll_node_down
        self.key_func["M-KEY_UP"] = self.scroll_node_up
        self.key_func["KEY_DOWN"] = self.item_listview.move_cursor_down
        self.key_func["KEY_UP"] = self.item_listview.move_cursor_up
        self.key_func["^M"] = self.open_selected_item
        self.resize()

        self.get_nodes()

    def command_publish_item(self):
        self.core.background = True
        editor = config.get('editor', '') or os.getenv('EDITOR') or 'vi'
        log.debug('Starting item edition with command %s' % editor)
        tmpfile = tempfile.NamedTemporaryFile(mode='r+')
        tmpfile.write(item_template % {'author': self.core.xmpp.boundjid.user, 'jid': self.core.xmpp.boundjid.bare})
        tmpfile.flush()
        process = subprocess.call(editor.split() + [tmpfile.name])
        tmpfile.flush()
        tmpfile.seek(0)
        item_dict = parse_template(tmpfile.readlines())
        tmpfile.close()
        log.debug('[%s]' % item_dict)

        self.core.background = False
        self.core.full_screen_redraw()
        entry = create_entry_from_dict(item_dict)
        self.publish_item(entry, self.get_selected_node_name())

    def publish_item(self, content, node_name):
        """
        publish the given item on the given node
        """
        def callback(res):
            if res:
                self.set_info_message('Item published')
            else:
                self.set_info_message('Item not published')
            self.force_refresh()

        self.core.xmpp.plugin['xep_0060'].setItem(self.server, node_name, content, callback=callback)

    def set_info_message(self, message):
        """
        Set an informative message in the upper line, near the server name
        """
        self.upper_message.set_message('Pubsub server: %s/%s [%s]' % (self.server, self.current_node or '', message))

    def resize(self):
        self.upper_message.resize(1, self.width, 0, 0)
        self.tab_win.resize(1, self.width, self.height-2, 0)

        column_size = {'node': self.width//4,
                       'name': self.width//4,}
        self.node_list_header.resize_columns(column_size)
        self.node_list_header.resize(1, self.width//2, 1, 0)
        self.node_listview.resize_columns(column_size)
        self.node_listview.resize(self.height//2-2, self.width//2, 2, 0)

        w = self.width//2
        column_size = {'id': w//8,
                       'title':w//8*5,
                       'author':w//8*2}
        self.item_list_header.resize_columns(column_size)
        self.item_list_header.resize(self.height//2+1, self.width//2, self.height//2, 0)
        self.item_listview.resize_columns(column_size)
        self.item_listview.resize(self.height//2-3, self.width//2, self.height//2+1, 0)

        self.vertical_separator.resize(self.height-3, 1, 1, self.width//2)

        self.item_viewer.resize(self.height-3, self.width//2+1, 1, self.width//2+1)
        self.input.resize(1, self.width, self.height-1, 0)

    def refresh(self):
        if self.need_resize:
            self.resize()
        log.debug('  TAB   Refresh: %s'%self.__class__.__name__)
        self.upper_message.refresh()
        self.node_list_header.refresh()
        self.node_listview.refresh()
        self.item_list_header.refresh()
        self.item_listview.refresh()
        self.item_viewer.refresh()
        self.tab_win.refresh()
        self.vertical_separator.refresh()
        self.input.refresh()

    def force_refresh(self):
        if self.core.current_tab() is self:
            self.core.refresh_window()

    def get_name(self):
        return '%s@@pubsubbrowser' % (self.server,)

    def on_input(self, key):
        res = self.input.do_command(key)
        if res:
            return True
        if key in self.key_func:
            return self.key_func[key]()

    def get_selected_node_name(self):
        """
        From the node_view_list, returns the node name of the selected
        one. None can be returned
        """
        line = self.node_listview.get_selected_row()
        if not line:
            return None
        return line['node']

    def get_node_by_name(self, name):
        """
        in the currently browsed node (or on the root), return the node with that name
        """
        nodes = self.current_node and self.current_node.subnodes or self.nodes
        for node in nodes:
            if node.name == name:
                return node
        return None

    def get_item_by_id(self, idd):
        """
        in the currently selected node, return the item with that id
        """
        selected_node_name = self.get_selected_node_name()
        if not selected_node_name:
            return None
        selected_node = self.get_node_by_name(selected_node_name)
        if not selected_node:
            return None
        for item in selected_node.items:
            if item.id == idd:
                return item
        return None

    def get_selected_item_id(self):
        """
        returns the id of the currently selected item
        """
        line = self.item_listview.get_selected_row()
        if not line:
            return None
        return line['id']

    def get_items(self, node):
        """
        Get all items in the given node
        """
        def callback(items):
            item_list = []
            if items:
                for it in items:
                    item_list.append(PubsubItem(it.attrib['id'], it))
                node.items = item_list
                log.debug('get_selected_node_name: %s' % self.get_selected_node_name())
                if self.get_selected_node_name() == node.name:
                    self.display_items_from_node(node)
            log.debug('Item on node %s: %s' % (node.name, item_list))
            self.set_info_message('Items received')
            self.force_refresh()

        self.core.xmpp.plugin['xep_0060'].get_items(self.server, node.name, callback=callback)

    def display_items_from_node(self, node):
        """
        takes a node, and set fill the item_listview with that
        node’s items
        """
        columns = self.item_list_header.get_columns()
        self.item_listview.empty()
        log.debug('display_items_from_node: %s' % node.items)
        for item in node.items:
            line = item.to_dict(columns)
            self.item_listview.lines.append(item.to_dict(columns))

    def add_nodes(self, node_list, parent=None):
        """
        Add Node objects to the list of the parent.
        If parent is None, they are added to the root list.
        If the current selected node is parent, we add
        them directly to the node_listview
        """
        log.debug('Adding nodes to %s: %s' % (node_list, parent,))
        if not parent:
            list_to_append = self.nodes
        else:
            list_to_append = parent.nodes
        self.node_listview.add_lines(node_list)
        for node in node_list:
            new_node = LeafNode(node['node'])
            list_to_append.append(new_node)
            self.get_items(new_node)

    def get_nodes(self, node=None):
        """
        Get all subnodes of the given node. If no node is given, get
        the root nodes
        """
        def callback(nodes):
            lines = [{'name': nodes[node] or '',
                  'node': node} for node in nodes.keys()]
            self.add_nodes(lines)
            self.set_info_message('Nodes received')
            self.force_refresh()
        self.core.xmpp.plugin['xep_0060'].get_nodes(self.server, callback=callback)

    def create_node(self, node_name):
        def callback(res):
            if res:
                self.node_listview.add_lines([{'name': '', 'node': node_name}])
                self.set_info_message('Node created')
            else:
                self.set_info_message('Node not created')
            self.reset_help_message()
            self.force_refresh()
        if node_name:
            self.core.xmpp.plugin['xep_0060'].create_node(self.server, node_name, callback=callback)
        return True

    def reset_help_message(self, txt=None):
        """
        Just reset the help message when a command ends
        """
        curses.curs_set(0)
        self.input = self.default_help_message
        return True

    def command_create_node(self):
        """
        Prompt for a node name and create it on Enter key
        """
        curses.curs_set(1)
        self.input = windows.CommandInput("[Create node]", self.reset_help_message, self.create_node, None)
        self.input.resize(1, self.width, self.height-1, 0)
        return True

    def scroll_node_up(self):
        """
        scroll the node up, and update the item list if needed
        """
        selected_node_before = self.get_selected_node_name()
        self.node_listview.move_cursor_up()
        selected_node_after = self.get_selected_node_name()
        if selected_node_after is not selected_node_before:
            self.display_items_from_node(self.get_node_by_name(selected_node_after))
            return True
        return False

    def scroll_node_down(self):
        """
        scroll the node down, and update the item list if needed
        """
        selected_node_before = self.get_selected_node_name()
        self.node_listview.move_cursor_down()
        selected_node_after = self.get_selected_node_name()
        if selected_node_after is not selected_node_before:
            self.display_items_from_node(self.get_node_by_name(selected_node_after))
            return True
        return False

    def open_selected_item(self):
        """
        displays the currently selected item in the item view window
        """
        selected_item = self.get_item_by_id(self.get_selected_item_id())
        if not selected_item:
            return
        log.debug('Content: %s'%ET.tostring(selected_item.content))
        entry = atom_parser.parse_atom_entry(selected_item.content)
        if not entry:
            self.item_viewer._text = str(ET.tostring(selected_item.content))
        else:
            self.item_viewer._text = \
"""\x193Title:\x19o %(title)s
\x193Author:\x19o %(author_name)s (%(author_uri)s)
%(dates)s\x193Link:\x19o %(link)s

\x193Summary:\x19o
%(summary)s
""" % {'title': entry.get('title') or '',
       'author_name': entry['author'].get('name') or '',
       'author_uri': entry['author'].get('uri') or '',
       'link': entry.get('link_href') or '',
       'summary': entry.get('summary') or '',
       'dates': '\x193Published:\x19o %(published)s\n%(updated)s' % {'published':entry.get('published') or '',
                                                                'updated': '' if (entry.get('updated') is None) or (entry.get('published') == entry.get('updated')) else '\x193Published:\x19o %s\n' % entry.get('updated')}
       }
        self.item_viewer.rebuild_text()
        return True



from sleekxmpp.xmlstream import ET
"""
Module used to provide fixes for sleekxmpp functions not yet fixed
upstream.

TODO: Check that they are fixed and remove those hacks
"""


def get_version(xmpp, jid, callback=None, **kwargs):
    def handle_result(res):
        if res and res['type'] != 'error':
            ret = res['software_version'].values
        else:
            ret = False
        if callback:
            callback(ret)
        return ret
    iq = xmpp.make_iq_get(ito=jid)
    iq['query'] = 'jabber:iq:version'
    result = iq.send(callback=handle_result if callback else None)
    if not callback:
        return handle_result(result)


def get_room_form(xmpp, room):
    iq = xmpp.make_iq_get(ito=room)
    query = ET.Element('{http://jabber.org/protocol/muc#owner}query')
    iq.append(query)
    try:
        result = iq.send()
    except:
        return False
    xform = result.xml.find('{http://jabber.org/protocol/muc#owner}query/{jabber:x:data}x')
    if xform is None:
        return False
    form = xmpp.plugin['xep_0004'].buildForm(xform)
    return form






"""
Installation
------------
You only have to load the plugin:

``/load alias``

Usage
-----

This plugin defines two new global commands: :term:`/alias` and :term:`/unalias`.

.. glossary::

    /alias
        **Usage:** ``/alias <name> <command> [args]``

        This command will create a new command, named ``name`` (and callable with
        ``/name``), that runs ``/command``, with ``[args]`` as fixed args for the command.
        When you run the alias, you can also pass parameters to it, that will be fed
        to the original command.

        Example: ::

            /alias toto say "koin "

        Will bind ``/say koin `` to ``/toto``, so this alias will work in any
        Chat tab. If someone calls it with ::

            /toto koin

        Poezio will then execute ``/say koin koin``.

    /unalias
        **Usage:** ``/unalias <name>``

        This command removes a defined alias.

"""

from plugin import BasePlugin
import common
from common import shell_split

class Plugin(BasePlugin):
    def init(self):
        self.api.add_command('alias', self.command_alias,
                usage='<alias> <command> [args]',
                short='Create an alias command',
                help='Create an alias for <command> with [args].')
        self.api.add_command('unalias', self.command_unalias,
                usage='<alias>',
                help='Remove a previously created alias',
                short='Remove an alias',
                completion=self.completion_unalias)
        self.commands = {}

    def command_alias(self, line):
        """
        /alias <alias> <command> [args]
        """
        arg = common.shell_split(line)
        if len(arg) < 2:
            self.api.information('Alias: Not enough parameters', 'Error')
            return
        alias = arg[0]
        command = arg[1]
        tmp_args = arg[2] if len(arg) > 2 else ''

        if alias in self.core.commands or alias in self.commands:
            self.api.information('Alias: command already exists', 'Error')
            return

        self.commands[alias] = lambda arg: self.get_command(command)(tmp_args + arg)
        self.api.add_command(alias, self.commands[alias], 'This command is an alias for /%s %s' %( command, tmp_args))
        self.api.information('Alias /%s successfuly created' % alias, 'Info')

    def command_unalias(self, alias):
        """
        /unalias <existing alias>
        """
        if alias in self.commands:
            del self.commands[alias]
            self.api.del_command(alias)
            self.api.information('Alias /%s successfuly deleted' % alias, 'Info')

    def completion_unalias(self, the_input):
        aliases = [alias for alias in self.commands]
        aliases.sort()
        return the_input.auto_completion(aliases, '', quotify=False)

    def get_command(self, name):
        """Returns the function associated with a command"""
        def dummy(args):
            """Dummy function called if the command doesn’t exist"""
            pass
        if name in self.core.commands:
            return self.core.commands[name][0]
        elif name in self.api.current_tab().commands:
            return self.api.current_tab().commands[name][0]
        return dummy



"""
Opens links in a browser.

Installation
------------

First use case: local use
~~~~~~~~~~~~~~~~~~~~~~~~~
If you use poezio on your workstation, this is for you.
You only have to load the plugin: ::

    /load link

Second use case: remote use
~~~~~~~~~~~~~~~~~~~~~~~~~~~
If you use poezio through SSH, this is for you.

.. note:: Small explanation: Poezio will create a `Unix FIFO`_ and send the commands in,
            and you will have to run a dæmon locally with ssh, to get those commands.

First, set the :term:`exec_remote` option in the config file to ``true``. Then select
the directory you want to put the fifo in (default is the current
directory, :file:`./`), the :file:`poezio.fifo` file will be created there.

After that, load the plugin: ::

    /load link

And open a link with :term:`/link` (as described below), this will create the FIFO.

You need to grab poezio’s sources on your client computer, or at least the `daemon.py`_
file.

Finally, on your client computer, run the ssh command:

.. code-block:: bash

    ssh toto@example.org "cat ~/poezio/poezio.fifo" | python3 daemon.py

Usage
-----

.. glossary::

    /link
        **Usage:** ``/link [range]``

        This plugin adds a :term:`/link` command that will open the links in ``firefox``. If
        you want to use another browser, you can use the :term:`/set` command  to change the
        :term:`browser` option.


        :term:`/link` without argument will open the last link found
        in the current tab, if any is found. An optional range
        argument can be given, to select one or more links to
        open. Examples:
        ``/link 1`` is equivalent to ``/link``
        ``/link 3`` will open the third link found in the current tab,
        starting from the bottom.
        ``/link 1:5`` will open the last five links in the current tab
        ``/link :2`` will open the last two links

Options
-------

:term:`exec_remote`

    To execute the command on your client

.. glossary::

    browser
        Set the default browser started by the plugin

.. _Unix FIFO: https://en.wikipedia.org/wiki/Named_pipe
.. _daemon.py: http://dev.louiz.org/projects/poezio/repository/revisions/master/raw/src/daemon.py

"""
import re

from plugin import BasePlugin
from xhtml import clean_text
import common
import tabs

url_pattern = re.compile(r'\b(http[s]?://(?:\S+))\b', re.I|re.U)

class Plugin(BasePlugin):
    def init(self):
        for _class in (tabs.MucTab, tabs.PrivateTab, tabs.ConversationTab):
            self.api.add_tab_command(_class, 'link', self.command_link,
                    usage='[num]',
                    help='Opens the last link from the conversation into a browser.\nIf [num] is given, then it will open the num-th link displayed.',
                    short='Open links into a browser')

    def find_link(self, nb):
        messages = self.api.get_conversation_messages()
        if not messages:
            return None
        for message in messages[::-1]:
            matches = url_pattern.findall(clean_text(message.txt))
            if matches:
                for url in matches[::-1]:
                    if nb == 1:
                        return url
                    else:
                        nb -= 1
        return None

    def command_link(self, args):
        args = common.shell_split(args)
        if len(args) == 1:
            if args[0].find(':') == -1:
                try:
                    start = int(args[0])
                    end = start
                except ValueError:
                    return self.api.run_command('/help link')
            else:
                start, end = args[0].split(':', 1)
                if start == '':
                    start = 1
                try:
                    start = int(start)
                    end = int(end)
                except ValueError:
                    return self.api.information('Invalid range: %s' % (args[0]), 'Error')
        else:
            start = 1
            end = 1
        for nb in range(start, end+1):
            link = self.find_link(nb)
            if not link:
                return self.api.information('No URL found.', 'Warning')
            self.core.exec_command([self.config.get('browser', 'firefox'), link])

    def cleanup(self):
        del self.config



"""
This plugin implements the `XEP-0027`_ “Current Jabber OpenPGP Usage”.

This is a plugin used to encrypt one-to-one conversation using the PGP
encryption method. You can use it if you want really good privacy. Without this
encryption, your messages are encrypted **at least** from your client (poezio) to
your server. The message is decrypted by your server and you cannot control the
encryption method of your messages from your server to your contact’s server
(unless you are your own server’s administrator), nor from your contact’s
server to your contact’s client.

This plugin does end-to-end encryption. This means that **only** your contact can
decrypt your messages, and it is fully encrypted during **all** its travel
through the internet.

Note that if you are having an encrypted conversation with a contact, you can
**not** send XHTML-IM messages to him. They will be removed and be replaced by
plain text messages.

Installation and configuration
------------------------------

You should autoload this plugin, as it will send your signed presence directly
on login, making it easier for your contact’s clients to know that you are
supporting GPG encryption. To do that, use the :term:`plugins_autoload` configuration
option.

You need to create a plugin configuration file. Create a file named :file:`gpg.cfg`
into your plugins configuration directory (:file:`~/.config/poezio/plugins` by
default), and fill it like this:

.. code-block:: ini

    [gpg]
    keyid = 091F9C78
    passphrase = your OPTIONAL passphrase

    [keys]
    example@jabber.org = E3CFCDE2
    juliet@xmpp.org = EF27ABCD

The ``gpg`` section is about your key. You need to specify the keyid, for the
key you want to use. You can as well provide a passphrase. If you don’t, you
should use a gpg agent or something like that that will ask your passphrase
whenever you need it.

The ``keys`` section contains your contact’s id keys. For each contact you want
to have encrypted conversations with, add her/his JID associated with the keyid
of his/her key.

And that’s it, now you need to talk directly to the **full** jid of your
contacts. Poezio doesn’t let you encrypt messages whom recipients is a bare
JID.

Additionnal information on GnuPG
--------------------------------

Create a key
~~~~~~~~~~~~

To create a personal key, use

.. code-block:: bash

    gpg --gen-key

and follow the instructions.

Keyid
~~~~~
The keyid (required in the gpg.cfg configuration file) is a 8 character-long
key. You can get the ones you created or imported by using the command

.. code-block:: bash

    gpg --list-keys

You will get something like

.. code-block:: none

    pub   4096R/01234567 2011-11-11
    uid                  Your Name Here (comment) <email@example.org>
    sub   4096R/AAFFBBCC 2011-11-11

    pub   2048R/12345678 2011-11-12 [expire: 2011-11-22]
    uid                  A contact’s name (comment) <fake@fake.fr>
    sub   2048R/FFBBAACC 2011-11-12 [expire: 2011-11-22]

In this example, the keyids are ``01234567`` and  ``12345678``.

Share your key
~~~~~~~~~~~~~~
Use:

.. code-block:: bash

    gpg --send-keys --keyserver pgp.mit.edu <keyid>

to upload you public key on a public server.

.. _XEP-0027: http://xmpp.org/extensions/xep-0027.html

"""
from gpg import gnupg
from sleekxmpp.xmlstream.stanzabase import JID

from xml.etree import cElementTree as ET
import xml.sax.saxutils

import logging
log = logging.getLogger(__name__)

from plugin import BasePlugin

from tabs import ConversationTab

NS_SIGNED = "jabber:x:signed"
NS_ENCRYPTED = "jabber:x:encrypted"


SIGNED_ATTACHED_MESSAGE = """-----BEGIN PGP SIGNED MESSAGE-----
Hash: %(hash)s

%(clear)s
-----BEGIN PGP SIGNATURE-----
Version: GnuPG

%(data)s
-----END PGP SIGNATURE-----
"""


ENCRYPTED_MESSAGE = """-----BEGIN PGP MESSAGE-----
Version: GnuPG

%(data)s
-----END PGP MESSAGE-----"""


class Plugin(BasePlugin):
    def init(self):
        self.contacts = {}
        # a dict of {full-JID: 'signed'/'valid'/'invalid'/'disabled'}
        # Whenever we receive a signed presence from a JID, we add it to this
        # dict, this way we know if we can encrypt the messages we will send to
        # this JID.
        # If that resource sends a non-signed presence, then we remove it
        # from that dict and stop encrypting our messages.
        # 'disabled' means that the user do NOT want to encrypt its messages
        # even if the key is valid.
        self.gpg = gnupg.GPG()
        self.keyid = self.config.get('keyid', '') or None
        self.passphrase = self.config.get('passphrase', '') or None
        if not self.keyid:
            self.core.information('No GPG keyid provided in the configuration', 'Warning')

        self.add_event_handler('send_normal_presence', self.sign_presence)
        self.add_event_handler('normal_presence', self.on_normal_presence)
        self.add_event_handler('conversation_say_after', self.on_conversation_say)
        self.add_event_handler('conversation_msg', self.on_conversation_msg)

        self.add_tab_command(ConversationTab, 'gpg', self.command_gpg,
                usage='<force|disable|setkey> [jid] [keyid]',
                help='Force or disable gpg encryption with the fulljid of the current conversation. The setkey argument lets you associate a keyid with the given bare JID.',
                short='Manage the GPG status',
                completion=self.gpg_completion)
        ConversationTab.add_information_element('gpg', self.display_encryption_status)

    def cleanup(self):
        self.send_unsigned_presence()
        ConversationTab.remove_information_element('gpg')
        self.del_tab_command(ConversationTab, 'gpg')

    def sign_presence(self, presence):
        """
        Sign every normal presence we send
        """
        signed_element = ET.Element('{%s}x' % (NS_SIGNED,))
        t = self.gpg.sign(presence['status'], keyid=self.keyid, passphrase=self.passphrase, detach=True)
        if not t:
            self.core.information('Could not sign presence. Disabling GPG module', 'Info')
            self.core.plugin_manager.unload('gpg')
            return
        text = xml.sax.saxutils.escape(str(t))
        signed_element.text = self.remove_gpg_headers(text)
        presence.append(signed_element)

    def send_unsigned_presence(self):
        """
        Send our current presence, to everyone, but unsigned, to indicate
        that we cannot/do not want to encrypt/decrypt messages.
        """
        current_presence = self.core.get_status()
        self.core.command_status('%s %s' % (current_presence.show or 'available', current_presence.message or '',))

    def on_normal_presence(self, presence, resource):
        """
        Check if it’s signed, if it is and we can verify the signature,
        add 'valid' or 'invalid' into the dict. If it cannot be verified, just add
        'signed'. Otherwise, do nothing.
        """
        signed = presence.find('{%s}x' % (NS_SIGNED,))
        bare = presence['from'].bare
        full = presence['from'].full
        if signed is None:
            if bare in self.contacts.keys():
                del self.contacts[bare]
            return
        if self.config.has_section('keys') and bare in self.config.options('keys'):
            self.contacts[full] = 'invalid'
            for hash_ in ('SHA1', 'SHA256'):
                to_verify = SIGNED_ATTACHED_MESSAGE % {'clear': presence['status'],
                                                       'data': signed.text,
                                                       'hash': hash_}
                verify = self.gpg.verify(to_verify)
                if verify:
                    self.contacts[full] = 'valid'
                    break
        else:
            self.contacts[full] = 'signed'

    def on_conversation_say(self, message, tab):
        """
        Check if the contact has a signed AND veryfied signature.
        If yes, encrypt the message with her key.
        """
        to = message['to']
        if not message['body']:
            # there’s nothing to encrypt if this is a chatstate, for example
            return
        signed = to.full in self.contacts.keys()
        if signed:
            veryfied = self.contacts[to.full] in ('valid', 'forced')
        else:
            veryfied = False
        if veryfied:
            # remove the xhtm_im body if present, because that
            # cannot be encrypted.
            del message['html']
            encrypted_element = ET.Element('{%s}x' % (NS_ENCRYPTED,))
            text = self.gpg.encrypt(message['body'], self.config.get(to.bare, '', section='keys'), always_trust=True)
            if not text:
                self.core.information('Could not encrypt message to %s' % (to.full),)
                # If we could not encrypt the message, don't send anything
                message['body'] = ''
                return
            encrypted_element.text = self.remove_gpg_headers(xml.sax.saxutils.escape(str(text)))
            message.append(encrypted_element)
            message['body'] = 'This message has been encrypted using the GPG key with id: %s' % self.keyid

    def on_conversation_msg(self, message, tab):
        """
        Check if the message is encrypted, and decrypt it if we can.
        """
        encrypted = message.find('{%s}x' % (NS_ENCRYPTED,))
        fro = message['from']
        if encrypted is not None:
            if self.config.has_section('keys') and fro.bare in self.config.options('keys'):
                keyid = self.config.get(fro.bare, '', 'keys')
                decrypted = self.gpg.decrypt(ENCRYPTED_MESSAGE % {'data': str(encrypted.text)}, passphrase=self.passphrase)
                if not decrypted:
                    self.core.information('Could not decrypt message from %s' % (fro.full),)
                    return
                message['body'] = str(decrypted)

    def display_encryption_status(self, jid):
        """
        Returns the status of encryption for the associated jid. This is to be used
        in the ConversationTab’s InfoWin.
        """
        if jid.full not in self.contacts.keys():
            return ''
        status = self.contacts[jid.full]
        if status in ('valid', 'invalid', 'signed'):
            return ' GPG Key: %s (%s)' % (status, 'encrypted' if status == 'valid' else 'NOT encrypted',)
        else:
            return ' GPG:  Encryption %s' % (status,)

    def command_gpg(self, args):
        """
        A command to force or disable the encryption, or to assign a keyid to a JID
        """
        args = args.split()
        if not args:
            return self.core.command_help("gpg")
        if len(args) >= 2:
            jid = JID(args[1])
        else:
            if isinstance(self.core.current_tab(), ConversationTab):
                jid = JID(self.core.current_tab().get_name())
            else:
                return
        command = args[0]
        if command == 'force' or command == 'enable':
            # we can force encryption only with contact having an associated
            # key, otherwise we cannot encrypt at all
            if self.config.has_section('keys') and jid.bare in self.config.options('keys'):
                self.contacts[JID(jid).full] = 'forced'
            else:
                self.core.information('Cannot force encryption: no key associated with %s' % (jid.bare), 'Info')
        elif command == 'disable':
            self.contacts[JID(jid).full] = 'disabled'
        elif command == 'setkey':
            if len(args) != 3:
                return self.core.command_help("gpg")
            if not self.config.has_section('keys'):
                self.config.add_section('keys')
            self.config.set(jid.bare, args[2], 'keys')
            self.config.write()
        self.core.refresh_window()

    def gpg_completion(self, the_input):
        return the_input.auto_completion(['force', 'disable', 'setkey'], ' ')

    def remove_gpg_headers(self, text):
        lines = text.splitlines()
        while lines[0].strip() != '':
            lines.pop(0)
        while lines[0].strip() == '':
            lines.pop(0)
        res = []
        for line in lines:
            if not line.startswith('---'):
                res.append(line)
        return '\n'.join(res)



# Copyright 2011 Florent Le Coz <louiz@louiz.org>
#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the zlib license. See the COPYING file.

"""
Defines the Fifo class
"""

import logging
log = logging.getLogger(__name__)

import os
import threading

class OpenTrick(threading.Thread):
    """
    A threaded trick to make the open for writing succeed.
    A fifo cannot be opened for writing if it has not been
    yet opened by the other hand for reading.
    So, we just open the fifo for reading and close it
    immediately afterwards.
    Once that is done, we can freely keep the fifo open for
    writing and write things in it. The writing can fail if
    there’s still nothing reading that fifo, but we just yell
    an error in that case.
    """
    def __init__(self, path):
        threading.Thread.__init__(self)
        self.path = path

    def run(self):
        open(self.path, 'r', encoding='utf-8').close()


class Fifo(object):
    """
    Just a simple file handler, writing and reading in a fifo.
    Mode is either 'r' or 'w', just like the mode for the open()
    function.
    """
    def __init__(self, path, mode):
        self.trick = None
        if not os.path.exists(path):
            os.mkfifo(path)
        if mode == 'w':
            self.trick = OpenTrick(path)
            # that thread will wait until we open it for writing
            self.trick.start()
        self.fd = open(path, mode, encoding='utf-8')

    def write(self, data):
        """
        Try to write on the fifo. If that fails, this means
        that nothing has that fifo opened, so the writing is useless,
        so we just return (and display an error telling that, somewhere).
        """
        self.fd.write(data)
        self.fd.flush()

    def readline(self):
        return self.fd.readline()

    def __del__(self):
        try:
            self.fd.close()
        except:
            pass



# Copyright 2010-2011 Florent Le Coz <louiz@louiz.org>
#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the zlib license. See the COPYING file.

from os import environ, makedirs
import mmap
import os
import re
from datetime import datetime
from config import config
from xhtml import clean_text
from theming import dump_tuple, get_theme

import logging

log = logging.getLogger(__name__)

from config import LOG_DIR

log_dir = os.path.join(LOG_DIR, 'logs')

message_log_re = re.compile('MR (\d{4})(\d{2})(\d{2})T(\d{2}):(\d{2}):(\d{2})Z (\d+) <([^ ]+)>  (.*)')
info_log_re = re.compile('MI (\d{4})(\d{2})(\d{2})T(\d{2}):(\d{2}):(\d{2})Z (\d+) (.*)')

def parse_message_line(msg):
    if re.match(message_log_re, msg):
        return [i for i in re.split(message_log_re, msg) if i]
    elif re.match(info_log_re, msg):
        return [i for i in re.split(info_log_re, msg) if i]
    return False


class Logger(object):
    """
    Appends things to files. Error/information/warning logs
    and also log the conversations to logfiles
    """
    def __init__(self):
        self.logfile = config.get('logfile', 'logs')
        self.roster_logfile = None
        # a dict of 'groupchatname': file-object (opened)
        self.fds = dict()

    def __del__(self):
        for opened_file in self.fds.values():
            if opened_file:
                try:
                    opened_file.close()
                except: # Can't close? too bad
                    pass

    def reload_all(self):
        """Close and reload all the file handles (on SIGHUP)"""
        for opened_file in self.fds.values():
            if opened_file:
                opened_file.close()
        log.debug('All log file handles closed')
        for room in self.fds:
            self.fds[room] = self.check_and_create_log_dir(room)
            log.debug('Log handle for %s re-created', room)

    def check_and_create_log_dir(self, room, open_fd=True):
        """
        Check that the directory where we want to log the messages
        exists. if not, create it
        """
        if config.get_by_tabname('use_log', 'true', room) == 'false':
            return
        try:
            makedirs(log_dir)
        except OSError as e:
            if e.errno != 17: # file exists
                log.error('Unable to create the log dir', exc_info=True)
        except:
            log.error('Unable to create the log dir', exc_info=True)
            return
        if not open_fd:
            return
        try:
            fd = open(os.path.join(log_dir, room), 'a')
            self.fds[room] = fd
            return fd
        except IOError:
            log.error('Unable to open the log file (%s)',
                    os.path.join(log_dir, room),
                    exc_info=True)

    def get_logs(self, jid, nb=10):
        """
        Get the nb last messages from the log history for the given jid.
        Note that a message may be more than one line in these files, so
        this function is a little bit more complicated than “read the last
        nb lines”.
        """
        if config.get_by_tabname('load_log', 10, jid) <= 0:
            return

        if config.get_by_tabname('use_log', 'true', jid) == 'false':
            return

        if nb <= 0:
            return

        self.check_and_create_log_dir(jid, open_fd=False)

        try:
            fd = open(os.path.join(log_dir, jid), 'rb')
        except:
            log.error('Unable to open the log file (%s)',
                    os.path.join(log_dir, jid),
                    exc_info=True)
            return
        if not fd:
            return

        # read the needed data from the file, we just search nb messages by
        # searching "\nM" nb times from the end of the file.  We use mmap to
        # do that efficiently, instead of seek()s and read()s which are costly.
        with fd:
            try:
                m = mmap.mmap(fd.fileno(), 0, prot=mmap.PROT_READ)
            except Exception: # file probably empty
                log.error('Unable to mmap the log file for (%s)',
                        os.path.join(log_dir, jid),
                        exc_info=True)
                return
            pos = m.rfind(b"\nM") # start of messages begin with MI or MR,
                                  # after a \n
            # number of message found so far
            count = 0
            while pos != -1 and count < nb-1:
                count += 1
                pos = m.rfind(b"\nM", 0, pos)
            if pos == -1:       # If we don't have enough lines in the file
                pos = 1         # 1, because we do -1 just on the next line
                                # to get 0 (start of the file)
            lines = m[pos-1:].decode(errors='replace').splitlines()

        messages = []
        color = '\x19%s}' % dump_tuple(get_theme().COLOR_INFORMATION_TEXT)

        # now convert that data into actual Message objects
        idx = 0
        while idx < len(lines):
            if lines[idx].startswith(' '): # should not happen ; skip
                idx += 1
                log.debug('fail?')
                continue
            tup = parse_message_line(lines[idx])
            idx += 1
            if not tup or 7 > len(tup) > 10  : # skip
                log.debug('format? %s', tup)
                continue
            time = [int(i) for index, i in enumerate(tup) if index < 6]
            message = {'lines': [], 'history': True, 'time': datetime(*time)}
            size = int(tup[6])
            if len(tup) == 8: #info line
                message['lines'].append(color+tup[7])
            else: # message line
                message['nickname'] = tup[7]
                message['lines'].append(color+tup[8])
            while size != 0 and idx < len(lines):
                message['lines'].append(lines[idx][1:])
                size -= 1
                idx += 1
            message['txt'] = '\n'.join(message['lines'])
            del message['lines']
            messages.append(message)

        return messages

    def log_message(self, jid, nick, msg, date=None, typ=1):
        """
        log the message in the appropriate jid's file
        type:
              0 = Don’t log
              1 = Message
              2 = Status/whatever
        """
        if not typ:
            return True

        jid = str(jid).replace('/', '\\')
        if config.get_by_tabname('use_log', 'false', jid) != 'true':
            return True
        if jid in self.fds.keys():
            fd = self.fds[jid]
        else:
            fd = self.check_and_create_log_dir(jid)
        if not fd:
            return True
        try:
            msg = clean_text(msg)
            if date is None:
                str_time = datetime.now().strftime('%Y%m%dT%H:%M:%SZ')
            else:
                str_time = date.strftime('%Y%m%dT%H:%M:%SZ')
            if typ == 1:
                prefix = 'MR'
            else:
                prefix = 'MI'
            lines = msg.split('\n')
            first_line = lines.pop(0)
            nb_lines = str(len(lines)).zfill(3)

            if nick:
                nick = '<' + nick + '>'
                fd.write(' '.join((prefix, str_time, nb_lines, nick, ' '+first_line, '\n')))
            else:
                fd.write(' '.join((prefix, str_time, nb_lines, first_line, '\n')))
            for line in lines:
                fd.write(' %s\n' % line)
        except:
            log.error('Unable to write in the log file (%s)',
                    os.path.join(log_dir, jid),
                    exc_info=True)
            return False
        else:
            try:
                fd.flush()          # TODO do something better here?
            except:
                log.error('Unable to flush the log file (%s)',
                        os.path.join(log_dir, jid),
                        exc_info=True)
                return False
        return True

    def log_roster_change(self, jid, message):
        """
        Log a roster change
        """
        if config.get_by_tabname('use_log', 'false', jid) != 'true':
            return True
        self.check_and_create_log_dir('', open_fd=False)
        if not self.roster_logfile:
            try:
                self.roster_logfile = open(os.path.join(log_dir, 'roster.log'), 'a')
            except IOError:
                log.error('Unable to create the log file (%s)',
                        os.path.join(log_dir, 'roster.log'),
                        exc_info=True)
                return False
        try:
            str_time = datetime.now().strftime('%Y%m%dT%H:%M:%SZ')
            message = clean_text(message)
            lines = message.split('\n')
            first_line = lines.pop(0)
            nb_lines = str(len(lines)).zfill(3)
            self.roster_logfile.write('MI %s %s %s %s\n' % (str_time, nb_lines, jid, first_line))
            for line in lines:
                self.roster_logfile.write(' %s\n' % line)
            self.roster_logfile.flush()
        except:
            log.error('Unable to write in the log file (%s)',
                    os.path.join(log_dir, 'roster.log'),
                    exc_info=True)
            return False
        return True

logger = Logger()



#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the zlib license. See the COPYING file.

"""
Defines the EventHandler class.
The list of available events is here:
http://poezio.eu/doc/en/plugins.html#_poezio_events
"""

import logging
log = logging.getLogger(__name__)

class EventHandler(object):
    """
    A class keeping a list of possible events that are triggered
    by poezio. You (a plugin for example) can add an event handler
    associated with an event name, and whenever that event is triggered,
    the callback is called.
    """
    def __init__(self):
        self.events = {
            'highlight': [],
            'muc_say': [],
            'muc_say_after': [],
            'conversation_say': [],
            'conversation_say_after': [],
            'private_say': [],
            'private_say_after': [],
            'conversation_msg': [],
            'private_msg': [],
            'muc_msg': [],
            'conversation_chatstate': [],
            'muc_chatstate': [],
            'private_chatstate': [],
            'normal_presence': [],
            'muc_presence': [],
            'muc_join': [],
            'joining_muc': [],
            'changing_nick': [],
            'muc_kick': [],
            'muc_nickchange': [],
            'muc_ban': [],
            'send_normal_presence': [],
            'ignored_private': [],
            }

    def add_event_handler(self, name, callback, position=0):
        """
        Add a callback to a given event.
        Note that if that event name doesn’t exist, it just returns False.
        If it was successfully added, it returns True
        position: 0 means insert at the beginning, -1 means end
        """
        if name not in self.events:
            return False

        if position >= 0:
            self.events[name].insert(position, callback)
        else:
            self.events[name].append(callback)

        return True

    def trigger(self, name, *args, **kwargs):
        """
        Call all the callbacks associated to the given event name.
        """
        callbacks = self.events.get(name, None)
        if callbacks is None:
            log.debug('%s: No such event.', name)
            return
        for callback in callbacks:
            callback(*args, **kwargs)

    def del_event_handler(self, name, callback):
        """
        Remove the callback from the list of callbacks of the given event
        """
        if not name:
            for event in self.events:
                while callback in self.events[event]:
                    self.events[event].remove(callback)
            return True
        else:
            if callback in self.events[name]:
                self.events[name].remove(callback)




# Copyright 2010-2011 Florent Le Coz <louiz@louiz.org>
#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the zlib license. See the COPYING file.

"""
Defines the data-forms Tab and all the Windows for it.
"""

import logging
log = logging.getLogger(__name__)
import curses

from windows import g_lock
import windows
from tabs import Tab
from theming import to_curses_attr, get_theme

class DataFormsTab(Tab):
    """
    A tab contaning various window type, displaying
    a form that the user needs to fill.
    """
    plugin_commands = {}
    def __init__(self, form, on_cancel, on_send, kwargs):
        Tab.__init__(self)
        self._form = form
        self._on_cancel = on_cancel
        self._on_send = on_send
        self._kwargs = kwargs
        self.fields = []
        for field in self._form:
            self.fields.append(field)
        self.topic_win = windows.Topic()
        self.form_win = FormWin(form, self.height-4, self.width, 1, 0)
        self.help_win = windows.HelpText("Ctrl+Y: send form, Ctrl+G: cancel")
        self.help_win_dyn = windows.HelpText()
        self.key_func['KEY_UP'] = self.form_win.go_to_previous_input
        self.key_func['KEY_DOWN'] = self.form_win.go_to_next_input
        self.key_func['^G'] = self.on_cancel
        self.key_func['^Y'] = self.on_send
        self.resize()
        self.update_commands()

    def on_cancel(self):
        self._on_cancel(self._form)
        return True

    def on_send(self):
        self._form.reply()
        self.form_win.reply()
        self._on_send(self._form)
        return True

    def on_input(self, key, raw=False):
        if key in self.key_func:
            res = self.key_func[key]()
            if res:
                return res
            self.help_win_dyn.refresh(self.form_win.get_help_message())
            self.form_win.refresh_current_input()
        else:
            self.form_win.on_input(key)

    def resize(self):
        self.need_resize = False
        self.topic_win.resize(1, self.width, 0, 0)
        self.form_win.resize(self.height-3 - Tab.tab_win_height(), self.width, 1, 0)
        self.help_win.resize(1, self.width, self.height-1, 0)
        self.help_win_dyn.resize(1, self.width, self.height-2 - Tab.tab_win_height(), 0)
        self.lines = []

    def refresh(self):
        if self.need_resize:
            self.resize()
        self.topic_win.refresh(self._form['title'])
        self.refresh_tab_win()
        self.help_win.refresh()
        self.help_win_dyn.refresh(self.form_win.get_help_message())
        self.form_win.refresh()

class FieldInput(object):
    """
    All input type in a data form should inherite this class,
    in addition with windows.Input or any relevant class from the
    'windows' library.
    """
    def __init__(self, field):
        self._field = field
        self.color = get_theme().COLOR_NORMAL_TEXT

    def set_color(self, color):
        self.color = color
        self.refresh()

    def update_field_value(self, value):
        raise NotImplementedError

    def resize(self, height, width, y, x):
        self._resize(height, width, y, x)

    def is_dummy(self):
        return False

    def reply(self):
        """
        Set the correct response value in the field
        """
        raise NotImplementedError

    def get_help_message(self):
        """
        Should return a string explaining the keys of the input.
        Will be displayed at each refresh on a line at the bottom of the tab.
        """
        return ''

class ColoredLabel(windows.Win):
    def __init__(self, text):
        self.text = text
        self.color = get_theme().COLOR_NORMAL_TEXT
        windows.Win.__init__(self)

    def resize(self, height, width, y, x):
        self._resize(height, width, y, x)

    def set_color(self, color):
        self.color = color
        self.refresh()

    def refresh(self):
        with g_lock:
            self._win.erase()
            self._win.attron(to_curses_attr(self.color))
            self.addstr(0, 0, self.text)
            self._win.attroff(to_curses_attr(self.color))
            self._refresh()


class DummyInput(FieldInput, windows.Win):
    """
    Used for fields that do not require any input ('fixed')
    """
    def __init__(self, field):
        FieldInput.__init__(self, field)
        windows.Win.__init__(self)

    def do_command(self):
        return

    def refresh(self):
        return

    def is_dummy(self):
        return True

class ColoredLabel(windows.Win):
    def __init__(self, text):
        self.text = text
        self.color = get_theme().COLOR_NORMAL_TEXT
        windows.Win.__init__(self)

    def resize(self, height, width, y, x):
        self._resize(height, width, y, x)

    def set_color(self, color):
        self.color = color
        self.refresh()

    def refresh(self):
        with g_lock:
            self._win.erase()
            self._win.attron(to_curses_attr(self.color))
            self.addstr(0, 0, self.text)
            self._win.attroff(to_curses_attr(self.color))
            self._refresh()

class BooleanWin(FieldInput, windows.Win):
    def __init__(self, field):
        FieldInput.__init__(self, field)
        windows.Win.__init__(self)
        self.last_key = 'KEY_RIGHT'
        self.value = bool(field.getValue())

    def do_command(self, key):
        if key == 'KEY_LEFT' or key == 'KEY_RIGHT':
            self.value = not self.value
            self.last_key = key
        self.refresh()

    def refresh(self):
        with g_lock:
            self._win.erase()
            self._win.attron(to_curses_attr(self.color))
            self.addnstr(0, 0, ' '*(8), self.width)
            self.addstr(0, 2, "%s"%self.value)
            self.addstr(0, 8, '→')
            self.addstr(0, 0, '←')
            if self.last_key == 'KEY_RIGHT':
                self.addstr(0, 8, '')
            else:
                self.addstr(0, 0, '')
            self._win.attroff(to_curses_attr(self.color))
            self._refresh()

    def reply(self):
        self._field['label'] = ''
        self._field.setAnswer(self.value)

    def get_help_message(self):
        return '← and →: change the value between True and False'

class TextMultiWin(FieldInput, windows.Win):
    def __init__(self, field):
        FieldInput.__init__(self, field)
        windows.Win.__init__(self)
        self.options = field.getValue()
        self.val_pos = 0
        self.edition_input = None
        if not isinstance(self.options, list):
            if isinstance(self.options, str):
                self.options = [self.options]
            else:
                self.options = []
        self.options.append('')

    def do_command(self, key):
        if not self.edition_input:
            if key == 'KEY_LEFT':
                if self.val_pos > 0:
                    self.val_pos -= 1
            elif key == 'KEY_RIGHT':
                if self.val_pos < len(self.options)-1:
                    self.val_pos += 1
            elif key == '^M':
                self.edition_input = windows.Input()
                self.edition_input.color = self.color
                self.edition_input.resize(self.height, self.width, self.y, self.x)
                self.edition_input.text = self.options[self.val_pos]
                self.edition_input.key_end()
        else:
            if key == '^M':
                self.options[self.val_pos] = self.edition_input.get_text()
                if not self.options[self.val_pos] and self.val_pos != len(self.options) -1:
                    del self.options[self.val_pos]
                    if self.val_pos == len(self.options) -1:
                        self.val_pos -= 1
                self.edition_input = None
                if not self.options or self.options[-1] != '':
                    self.options.append('')
            else:
                self.edition_input.do_command(key)
        self.refresh()

    def refresh(self):
        if not self.edition_input:
            with g_lock:
                self._win.erase()
                self._win.attron(to_curses_attr(self.color))
                self.addnstr(0, 0, ' '*self.width, self.width)
                option = self.options[self.val_pos]
                self.addstr(0, self.width//2-len(option)//2, option)
                if self.val_pos > 0:
                    self.addstr(0, 0, '←')
                if self.val_pos < len(self.options)-1:
                    self.addstr(0, self.width-1, '→')
                self._win.attroff(to_curses_attr(self.color))
                self._refresh()
        else:
            self.edition_input.refresh()

    def reply(self):
        values = [val for val in self.options if val]
        self._field.setAnswer(values)

    def get_help_message(self):
        if not self.edition_input:
            help_msg = '← and →: browse the available entries. '
            if self.val_pos == len(self.options)-1:
                help_msg += 'Enter: add an entry'
            else:
                help_msg += 'Enter: edit this entry'
        else:
            help_msg = 'Enter: finish editing this entry.'
        return help_msg

class ListMultiWin(FieldInput, windows.Win):
    def __init__(self, field):
        FieldInput.__init__(self, field)
        windows.Win.__init__(self)
        values = field.getValue()
        self.options = [[option, True if option['value'] in values else False]\
                            for option in field.getOptions()]
        self.val_pos = 0

    def do_command(self, key):
        if key == 'KEY_LEFT':
            if self.val_pos > 0:
                self.val_pos -= 1
        elif key == 'KEY_RIGHT':
            if self.val_pos < len(self.options)-1:
                self.val_pos += 1
        elif key == ' ':
            self.options[self.val_pos][1] = not self.options[self.val_pos][1]
        else:
            return
        self.refresh()

    def refresh(self):
        with g_lock:
            self._win.erase()
            self._win.attron(to_curses_attr(self.color))
            self.addnstr(0, 0, ' '*self.width, self.width)
            if self.val_pos > 0:
                self.addstr(0, 0, '←')
            if self.val_pos < len(self.options)-1:
                self.addstr(0, self.width-1, '→')
            option = self.options[self.val_pos]
            self.addstr(0, self.width//2-len(option)//2, option[0]['label'])
            self.addstr(0, 2, '✔' if option[1] else '☐')
            self._win.attroff(to_curses_attr(self.color))
            self._refresh()

    def reply(self):
        self._field['label'] = ''
        self._field.delOptions()
        values = [option[0]['value'] for option in self.options if option[1] is True]
        self._field.setAnswer(values)

    def get_help_message(self):
        return '←, →: Switch between the value. Space: select or unselect a value'

class ListSingleWin(FieldInput, windows.Win):
    def __init__(self, field):
        FieldInput.__init__(self, field)
        windows.Win.__init__(self)
        # the option list never changes
        self.options = field.getOptions()
        # val_pos is the position of the currently selected option
        self.val_pos = 0
        for i, option in enumerate(self.options):
            if field.getValue() == option['value']:
                self.val_pos = i

    def do_command(self, key):
        if key == 'KEY_LEFT':
            if self.val_pos > 0:
                self.val_pos -= 1
        elif key == 'KEY_RIGHT':
            if self.val_pos < len(self.options)-1:
                self.val_pos += 1
        else:
            return
        self.refresh()

    def refresh(self):
        with g_lock:
            self._win.erase()
            self._win.attron(to_curses_attr(self.color))
            self.addnstr(0, 0, ' '*self.width, self.width)
            if self.val_pos > 0:
                self.addstr(0, 0, '←')
            if self.val_pos < len(self.options)-1:
                self.addstr(0, self.width-1, '→')
            option = self.options[self.val_pos]['label']
            self.addstr(0, self.width//2-len(option)//2, option)
            self._win.attroff(to_curses_attr(self.color))
            self._refresh()

    def reply(self):
        self._field['label'] = ''
        self._field.delOptions()
        self._field.setAnswer(self.options[self.val_pos]['value'])

    def get_help_message(self):
        return '←, →: Select a value amongst the others'

class TextSingleWin(FieldInput, windows.Input):
    def __init__(self, field):
        FieldInput.__init__(self, field)
        windows.Input.__init__(self)
        self.text = field.getValue() if isinstance(field.getValue(), str)\
            else ""
        self.pos = len(self.text)
        self.color = get_theme().COLOR_NORMAL_TEXT

    def reply(self):
        self._field['label'] = ''
        self._field.setAnswer(self.get_text())

    def get_help_message(self):
        return 'Edit the text'

class TextPrivateWin(TextSingleWin):
    def __init__(self, field):
        TextSingleWin.__init__(self, field)

    def rewrite_text(self):
        with g_lock:
            self._win.erase()
            if self.color:
                self._win.attron(to_curses_attr(self.color))
            self.addstr('*'*len(self.text[self.view_pos:self.view_pos+self.width-1]))
            if self.color:
                (y, x) = self._win.getyx()
                size = self.width-x
                self.addnstr(' '*size, size, to_curses_attr(self.color))
            self.addstr(0, self.pos, '')
            if self.color:
                self._win.attroff(to_curses_attr(self.color))
            self._refresh()

    def get_help_message(self):
        return 'Edit the secret text'

class FormWin(object):
    """
    A window, with some subwins (the various inputs).
    On init, create all the subwins.
    On resize, move and resize all the subwin and define how the text will be written
    On refresh, write all the text, and refresh all the subwins
    """
    input_classes = {'boolean': BooleanWin,
                     'fixed': DummyInput,
                     'jid-multi': TextMultiWin,
                     'jid-single': TextSingleWin,
                     'list-multi': ListMultiWin,
                     'list-single': ListSingleWin,
                     'text-multi': TextMultiWin,
                     'text-private': TextPrivateWin,
                     'text-single': TextSingleWin,
                     }
    def __init__(self, form, height, width, y, x):
        self._form = form
        self._win = windows.Win._tab_win.derwin(height, width, y, x)
        self.current_input = 0
        self.inputs = []        # dict list
        for (name, field) in self._form.getFields().items():
            if field['type'] == 'hidden':
                continue
            try:
                input_class = self.input_classes[field['type']]
            except IndexError:
                continue
            label = field['label']
            desc = field['desc']
            if field['type'] == 'fixed':
                label = field.getValue()
            inp = input_class(field)
            self.inputs.append({'label':ColoredLabel(label),
                                'description': desc,
                                'input':inp})

    def resize(self, height, width, y, x):
        self.height = height
        self.width = width
        if self.current_input >= self.height-2:
            self.current_input = self.height-1

    def reply(self):
        """
        Set the response values in the form, for each field
        from the corresponding input
        """
        for inp in self.inputs:
            if inp['input'].is_dummy():
                continue
            else:
                inp['input'].reply()
        self._form['title'] = ''
        self._form['instructions'] = ''

    def go_to_next_input(self):
        if not self.inputs:
            return
        if self.current_input == len(self.inputs) - 1 or self.current_input >= self.height-1:
            return
        self.inputs[self.current_input]['input'].set_color(get_theme().COLOR_NORMAL_TEXT)
        self.inputs[self.current_input]['label'].set_color(get_theme().COLOR_NORMAL_TEXT)
        self.current_input += 1
        jump = 0
        while self.current_input+jump != len(self.inputs) - 1 and self.inputs[self.current_input+jump]['input'].is_dummy():
            jump += 1
        if self.inputs[self.current_input+jump]['input'].is_dummy():
            return
        self.current_input += jump
        self.inputs[self.current_input]['input'].set_color(get_theme().COLOR_SELECTED_ROW)
        self.inputs[self.current_input]['label'].set_color(get_theme().COLOR_SELECTED_ROW)

    def go_to_previous_input(self):
        if not self.inputs:
            return
        if self.current_input == 0:
            return
        self.inputs[self.current_input]['input'].set_color(get_theme().COLOR_NORMAL_TEXT)
        self.inputs[self.current_input]['label'].set_color(get_theme().COLOR_NORMAL_TEXT)
        self.current_input -= 1
        jump = 0
        while self.current_input-jump > 0 and self.inputs[self.current_input+jump]['input'].is_dummy():
            jump += 1
        if self.inputs[self.current_input+jump]['input'].is_dummy():
            return
        self.current_input -= jump
        self.inputs[self.current_input]['input'].set_color(get_theme().COLOR_SELECTED_ROW)
        self.inputs[self.current_input]['label'].set_color(get_theme().COLOR_SELECTED_ROW)

    def on_input(self, key):
        if not self.inputs:
            return
        self.inputs[self.current_input]['input'].do_command(key)

    def refresh(self):
        with g_lock:
            self._win.erase()
            y = 0
            i = 0
            for name, field in self._form.getFields().items():
                if field['type'] == 'hidden':
                    continue
                self.inputs[i]['label'].resize(1, self.width//3, y + 1, 0)
                self.inputs[i]['input'].resize(1, self.width//3, y+1, 2*self.width//3)
                # TODO: display the field description
                y += 1
                i += 1
                if y >= self.height:
                    break
            self._win.refresh()
        for i, inp in enumerate(self.inputs):
            if i >= self.height:
                break
            inp['label'].refresh()
            inp['input'].refresh()
            inp['label'].refresh()
        if self.current_input < self.height-1:
            self.inputs[self.current_input]['input'].set_color(get_theme().COLOR_SELECTED_ROW)
            self.inputs[self.current_input]['input'].refresh()
            self.inputs[self.current_input]['label'].set_color(get_theme().COLOR_SELECTED_ROW)
            self.inputs[self.current_input]['label'].refresh()

    def refresh_current_input(self):
        self.inputs[self.current_input]['input'].refresh()

    def get_help_message(self):
        if self.current_input < self.height-1 and self.inputs[self.current_input]['input']:
            return self.inputs[self.current_input]['input'].get_help_message()
        return ''



"""
This plugin lets you execute a command, to notify you from new important
messages.

Installation and configuration
------------------------------

You need to create a plugin configuration file. Create a file named :file:`simple_notify.cfg`
into your plugins configuration directory (:file:`~/.config/poezio/plugins` by
default), and fill it like this:

First example:

.. code-block:: ini

    [simple_notify]
    command = notify-send -i /path/to/poezio/data/poezio_80.png "New message from %(from)s" "%(body)s"

Second example:

.. code-block:: ini

    [simple_notify]
    command = echo \\<%{from}s\\> %{body}s >> some.fifo
    delay = 3
    after_command = echo >> some.fifo

You can put any command, instead of these ones. You can also use the
special keywords ``%(from)s`` and ``%(body)s`` that will be replaced
directly in the command line by the author of the message, and the body.

The first example shown above will display something like this:

.. figure:: ../images/simple_notify_example.png
    :alt: Simple notify example

The second example will first write the author and the message in a
fifo, that fifo can locally be read by some other program (was tested
with the xmobar PipeReader command, which displays what is read from a
fifo into a status bar. Be careful, you have two different fifos in
that case, don’t get confused). The :term:`delay` and :term:`after_command` options
are used to erase/delete/kill the notification after a certain
delay.  In our example it is used to display an empty message in our
xmobar, erasing the notification after 3 seconds.

.. note:: If you set the :term:`exec_remote` option to ``true`` into the
    main configuration file, the command will be executed remotely
    (as explained in the :ref:`link-plugin` plugin help).

Options defined
---------------

.. glossary::
    :sorted:

    command
        The command to execute (with special keywords ``%{from}s`` and ``${body}s``)

    delay
        Delay after which :term:`after_command` must be executed.

    after_command
        Command to run after :term:`delay`. You probably want to clean up things.

"""

from plugin import BasePlugin
from xhtml import get_body_from_message_stanza
from timed_events import DelayedEvent
import shlex

class Plugin(BasePlugin):
    def init(self):
        self.api.add_event_handler('private_msg', self.on_private_msg)
        self.api.add_event_handler('conversation_msg', self.on_conversation_msg)
        self.api.add_event_handler('highlight', self.on_highlight)

    def on_private_msg(self, message, tab):
        fro = message['from']
        self.do_notify(message, fro)

    def on_highlight(self, message, tab):
        fro = message['from'].resource
        self.do_notify(message, fro)

    def on_conversation_msg(self, message, tab):
        fro = message['from'].bare
        self.do_notify(message, fro)

    def do_notify(self, message, fro):
        body = get_body_from_message_stanza(message, use_xhtml=False)
        if not body:
            return
        command_str = self.config.get('command', '').strip()
        if not command_str:
            self.api.information('No notification command was provided in the configuration file', 'Warning')
            return
        command = [arg % {'body': body.replace('\n', ' '), 'from': fro} for arg in shlex.split(command_str)]
        self.core.exec_command(command)
        after_command_str = self.config.get('after_command', '').strip()
        if not after_command_str:
            return
        after_command = [arg % {'body': body.replace('\n', ' '), 'from': fro} for arg in shlex.split(after_command_str)]
        delayed_event = DelayedEvent(self.config.get('delay', 1), self.core.exec_command, after_command)
        self.api.add_timed_event(delayed_event)



#!/usr/bin/env python3
from distutils.core import setup, Extension
import os, sys

module_poopt = Extension('poezio.poopt',
                    sources = ['src/pooptmodule.c'])


current_dir = os.path.dirname(__file__)

# Create a link to the config file (for packaging purposes)
if not os.path.exists(os.path.join(current_dir, 'src', 'default_config.cfg')):
    os.link(os.path.join(current_dir, 'data', 'default_config.cfg'),
            os.path.join(current_dir, 'src',  'default_config.cfg'))

setup(name="poezio",
       version="0.8",
       description="A console XMPP client",
       long_description=
       "Poezio is a Free chat client aiming to reproduce the ease of use of most "
       "IRC clients (e.g. weechat, irssi) while using the XMPP network."
       "\n"
       "Documentation is available at http://doc.poez.io/0.8.",


       ext_modules = [module_poopt],
       url = 'http://poez.io/',
       license = 'zlib',
       download_url = 'https://dev.louiz.org/projects/poezio/files',

       author = 'Florent Le Coz',
       author_email = 'louiz@louiz.org',

       maintainer = 'Mathieu Pasquet',
       maintainer_email = 'mathieui@mathieui.net',

       classifiers = [ 'Development Status :: 4 - Beta',
                       'Topic :: Communications :: Chat',
                       'Environment :: Console :: Curses',
                       'Intended Audience :: End Users/Desktop',
                       'License :: OSI Approved :: zlib/libpng License',
                       'Natural Language :: English',
                       'Operating System :: Unix',
                       'Programming Language :: Python :: 3'
                    ],
       keywords = ['jabber', 'xmpp', 'client', 'chat', 'im', 'console'],
       packages = ['poezio', 'poezio_plugins', 'poezio_plugins.gpg', 'poezio_themes'],
       package_dir = {'poezio': 'src', 'poezio_plugins': 'plugins', 'poezio_themes': 'data/themes'},
       package_data = {'poezio': ['default_config.cfg']},
       scripts = ['scripts/poezio'],
       data_files = [('share/man/man1/', ['data/poezio.1'])],

       install_requires = ['sleekxmpp==1.2.4',
                           'dnspython3>=1.11.1'],
       extras_require = {'OTR plugin': 'python-potr>=1.0',
                         'Screen autoaway plugin': 'pyinotify==0.9.4'}
)

# Remove the link afterwards
if os.path.exists(os.path.join(current_dir, 'src', 'default_config.cfg')):
    os.unlink(os.path.join(current_dir, 'src', 'default_config.cfg'))




# Copyright 2010-2011 Florent Le Coz <louiz@louiz.org>
#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the zlib license. See the COPYING file.

"""
Timed events are the standard way to schedule events for later in poezio.

Once created, they must be added to the list of checked events with
:py:func:`Core.add_timed_event` (within poezio) or with
:py:func:`.PluginAPI.add_timed_event` (within a plugin).
"""

import logging

log = logging.getLogger(__name__)

import datetime

class TimedEvent(object):
    """
    An event with a callback that is called when the specified time is passed.

    Note that these events can NOT be used for very small delay or a very
    precise date, since the check for events is done once per second, as
    a maximum.

    The callback and its arguments should be passed as the lasts arguments.
    """
    def __init__(self, date, callback, *args):
        """
        Create a new timed event.

        :param datetime.datetime date: Time at which the callback must be run.
        :param function callback: The handler that will be executed.
        :param \*args: Optional arguments passed to the handler.
        """
        self._callback = callback
        self.args = args
        self.repetive = False
        self.next_call_date = date

    def __call__(self):
        """
        the call should return False if this event should be remove from
        the events list.
        If it’s true, the date should be updated beforehand to a later date,
        or else it will be called every second
        """
        self._callback(*self.args)
        return self.repetive

    def has_timed_out(self, current_date):
        """
        Check if the event has timed out.

        :param datetime.datetime current_date: The current date.
        :returns: True if the callback should be called
        :rtype: bool
        """
        if self.next_call_date < current_date:
            return True
        else:
            return False

    def change_date(self, date):
        """
        Simply change the date of the event.

        :param datetime.datetime date: Next date.
        """
        self.next_call_date = date

    def add_delay(self, delay):
        """
        Add a delay (in seconds) to the date.

        :param int delay: The delay to add.
        """
        self.next_call_date += datetime.timedelta(seconds=delay)

class DelayedEvent(TimedEvent):
    """
    A TimedEvent, but with the date calculated from now + a delay in seconds.
    Use it if you want an event to happen in, e.g. 6 seconds.
    """
    def __init__(self, delay, callback, *args):
        """
        Create a new DelayedEvent.

        :param int delay: The number of seconds.
        :param function callback: The handler that will be executed.
        :param \*args: Optional arguments passed to the handler.
        """
        date = datetime.datetime.now() + datetime.timedelta(seconds=delay)
        TimedEvent.__init__(self, date, callback, *args)




"""
Module related to the argument parsing

There is a fallback to the deprecated optparse if argparse is not found
"""
from os import path

def parse_args(CONFIG_PATH=''):
    """
    Parse the arguments from the command line
    """
    try:
        from argparse import ArgumentParser, SUPPRESS
    except ImportError:
        from optparse import OptionParser
        from optparse import SUPPRESS_HELP as SUPPRESS
        parser = OptionParser()
        parser.add_option("-f", "--file", dest="filename", default=path.join(CONFIG_PATH, 'poezio.cfg'),
                            help="The config file you want to use", metavar="CONFIG_FILE")
        parser.add_option("-d", "--debug", dest="debug",
                            help="The file where debug will be written", metavar="DEBUG_FILE")
        parser.add_option("-v", "--version", dest="version",
                            help=SUPPRESS, metavar="VERSION", default="0.8-dev")
        (options, args) = parser.parse_args()
    else:
        parser = ArgumentParser()
        parser.add_argument("-f", "--file", dest="filename", default=path.join(CONFIG_PATH, 'poezio.cfg'),
                            help="The config file you want to use", metavar="CONFIG_FILE")
        parser.add_argument("-d", "--debug", dest="debug",
                            help="The file where debug will be written", metavar="DEBUG_FILE")
        parser.add_argument("-v", "--version", dest="version",
                            help=SUPPRESS, metavar="VERSION", default="0.8-dev")
        options = parser.parse_args()
    return options



# Copyright 2010-2011 Florent Le Coz <louiz@louiz.org>
#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the zlib license. See the COPYING file.

"""
Defines the global config instance, used to get or set (and save) values
from/to the config file
"""

DEFSECTION = "Poezio"

from gettext import gettext as _
import sys
import tempfile

import os
import logging

from configparser import RawConfigParser, NoOptionError, NoSectionError
from os import environ, makedirs, path, remove
from shutil import copy2
from args import parse_args


class Config(RawConfigParser):
    """
    load/save the config to a file
    """
    def __init__(self, file_name):
        RawConfigParser.__init__(self, None)
        # make the options case sensitive
        self.optionxform = str
        self.read_file(file_name)

    def read_file(self, file_name):
        self.file_name = file_name
        try:
            RawConfigParser.read(self, file_name, encoding='utf-8')
        except TypeError: # python < 3.2 sucks
            RawConfigParser.read(self, file_name)
        # Check config integrity and fix it if it’s wrong
        for section in ('bindings', 'var'):
            if not self.has_section(section):
                self.add_section(section)

    def get(self, option, default, section=DEFSECTION):
        """
        get a value from the config but return
        a default value if it is not found
        The type of default defines the type
        returned
        """
        try:
            if type(default) == int:
                res = self.getint(option, section)
            elif type(default) == float:
                res = self.getfloat(option, section)
            elif type(default) == bool:
                res = self.getboolean(option, section)
            else:
                res = self.getstr(option, section)
        except (NoOptionError, NoSectionError):
            return default
        return res

    def getl(self, option, default, section=DEFSECTION):
        """
        get a value and return it lowercase
        """
        return self.get(option, default, section).lower()

    def get_by_tabname(self, option, default, tabname, fallback=True, fallback_server=True):
        """
        Try to get the value for the option. First we look in
        a section named `tabname`, if the option is not present
        in the section, we search for the global option if fallback is
        True. And we return `default` as a fallback as a last resort.
        """
        if tabname in self.sections():
            if option in self.options(tabname):
                # We go the tab-specific option
                return self.get(option, default, tabname)
        if fallback_server:
            return self.get_by_servname(tabname, option, default, fallback)
        if fallback:
            # We fallback to the global option
            return self.get(option, default)
        return default

    def get_by_servname(self, jid, option, default, fallback=True):
        """
        Try to get the value of an option for a server
        """
        server = safeJID(jid).server
        if server:
            server = '@' + server
            if server in self.sections() and option in self.options(server):
                return self.get(option, default, server)
        if fallback:
            return self.get(option, default)
        return default


    def __get(self, option, section=DEFSECTION):
        """
        facility for RawConfigParser.get
        """
        return RawConfigParser.get(self, section, option)

    def getstr(self, option, section=DEFSECTION):
        """
        get a value and returns it as a string
        """
        return self.__get(option, section)

    def getint(self, option, section=DEFSECTION):
        """
        get a value and returns it as an int
        """
        try:
            return int(self.__get(option, section))
        except ValueError:
            return -1

    def getfloat(self, option, section=DEFSECTION):
        """
        get a value and returns it as a float
        """
        return float(self.__get(option, section))

    def getboolean(self, option, section=DEFSECTION):
        """
        get a value and returns it as a boolean
        """
        return RawConfigParser.getboolean(self, section, option)

    def write_in_file(self, section, option, value):
        """
        Our own way to save write the value in the file
        Just find the right section, and then find the
        right option, and edit it.

        TODO: make it write also new values in the file, not just what did already
        exist
        """
        if path.exists(self.file_name):
            df = open(self.file_name, 'r', encoding='utf-8')
            lines_before = (line.strip() for line in df.readlines())
            df.close()
        else:
            lines_before = []
        result_lines = []
        we_are_in_the_right_section = False
        written = False
        section_found = False
        for line in lines_before:
            if line.startswith('['): # check the section
                if we_are_in_the_right_section and not written:
                    result_lines.append('%s = %s' % (option, value))
                    written = True
                if line == '[%s]' % section:
                    we_are_in_the_right_section = True
                    section_found = True
                else:
                    we_are_in_the_right_section = False
            if (line.startswith('%s ' % (option,)) or
                line.startswith('%s=' % (option,)) or
                line.startswith('%s = ' % (option,))) and we_are_in_the_right_section:
                line = '%s = %s' % (option, value)
                written = True
            result_lines.append(line)

        if not section_found:
            result_lines.append('[%s]' % section)
            result_lines.append('%s = %s' % (option, value))
        elif not written:
            result_lines.append('%s = %s' % (option, value))
        try:
            prefix, file = path.split(self.file_name)
            filename = path.join(prefix, '.%s.tmp' % file)
            fd = os.fdopen(
                    os.open(
                        filename,
                        os.O_WRONLY | os.O_CREAT,
                        0o600),
                    'w')
            for line in result_lines:
                fd.write('%s\n' % line)
            fd.close()
            copy2(filename, self.file_name)
            remove(filename)
        except:
            success = False
            log.error('Unable to save the config file.', exc_info=True)
        else:
            success = True
        return success

    def set_and_save(self, option, value, section=DEFSECTION):
        """
        set the value in the configuration then save it
        to the file
        """
        # Special case for a 'toggle' value. We take the current value
        # and set the opposite. Warning if the no current value exists
        # or it is not a bool.
        if value == "toggle":
            current = self.get(option, "", section)
            if current.lower() == "false":
                value = "true"
            elif current.lower() == "true":
                value = "false"
            else:
                return (_("Could not toggle option: %s. Current value is %s.") % (option, current or _("empty")), 'Warning')
        if self.has_section(section):
            RawConfigParser.set(self, section, option, value)
        else:
            self.add_section(section)
            RawConfigParser.set(self, section, option, value)
        if not self.write_in_file(section, option, value):
            return (_('Unable to write in the config file'), 'Error')
        return ("%s=%s" % (option, value), 'Info')

    def silent_set(self, option, value, section=DEFSECTION):
        """
        Set a value, save, and return True on success and False on failure
        """
        if self.has_section(section):
            RawConfigParser.set(self, section, option, value)
        else:
            self.add_section(section)
            RawConfigParser.set(self, section, option, value)
        return self.write_in_file(section, option, value)

    def set(self, option, value, section=DEFSECTION):
        """
        Set the value of an option temporarily
        """
        try:
            RawConfigParser.set(self, section, option, value)
        except NoSectionError:
            pass

    def to_dict(self):
        """
        Returns a dict of the form {section: {option: value, option: value}, …}
        """
        res = {}
        for section in self.sections():
            res[section] = {}
            for option in self.options(section):
                res[section][option] = self.get(option, "", section)
        return res

firstrun = False

# creates the configuration directory if it doesn't exist
# and copy the default config in it
CONFIG_HOME = environ.get("XDG_CONFIG_HOME")
if not CONFIG_HOME:
    CONFIG_HOME = path.join(environ.get('HOME'), '.config')
CONFIG_PATH = path.join(CONFIG_HOME, 'poezio')

try:
    makedirs(CONFIG_PATH)
except OSError:
    pass

options = parse_args(CONFIG_PATH)

# Copy a default file if none exists
if not path.isfile(options.filename):
    default = path.join(path.dirname(__file__), '../data/default_config.cfg')
    other = path.join(path.dirname(__file__), 'default_config.cfg')
    if path.isfile(default):
        copy2(default, options.filename)
    elif path.isfile(other):
        copy2(other, options.filename)
    firstrun = True

try:
    config = Config(options.filename)
except:
    import traceback
    sys.stderr.write('Poezio was unable to read or parse the config file.\n')
    traceback.print_exc(limit=0)
    sys.exit(1)

LOG_DIR = config.get('log_dir', '') or path.join(environ.get('XDG_DATA_HOME') or path.join(environ.get('HOME'), '.local', 'share'), 'poezio')
LOG_DIR = path.expanduser(LOG_DIR)

try:
    makedirs(LOG_DIR)
except:
    pass

LOGGING_CONFIG = {
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
        'simple': {
            'format': '%(asctime)s %(levelname)s:%(module)s:%(message)s'
        }
    },
    'handlers': {
    },
    'root': {
            'handlers': [],
            'propagate': True,
            'level': 'DEBUG',
    }
}
if config.get('log_errors', 'true').lower() != 'false':
    LOGGING_CONFIG['root']['handlers'].append('error')
    LOGGING_CONFIG['handlers']['error'] = {
            'level': 'ERROR',
            'class': 'logging.FileHandler',
            'filename': path.join(LOG_DIR, 'errors.log'),
            'formatter': 'simple',
        }

if options.debug:
    LOGGING_CONFIG['root']['handlers'].append('debug')
    LOGGING_CONFIG['handlers']['debug'] = {
            'level':'DEBUG',
            'class':'logging.FileHandler',
            'filename': options.debug,
            'formatter': 'simple',
        }


if LOGGING_CONFIG['root']['handlers']:
    logging.config.dictConfig(LOGGING_CONFIG)
else:
    logging.basicConfig(level=logging.CRITICAL)

# common import sleekxmpp, which creates then its loggers, so
# it needs to be after logger configuration
from common import safeJID

log = logging.getLogger(__name__)




"""
Module containing various decorators
"""

class RefreshWrapper(object):
    def __init__(self):
        self.core = None

    def conditional(self, func):
        """
        Decorator to refresh the UI if the wrapped function
        returns True
        """
        def wrap(*args, **kwargs):
            ret = func(*args, **kwargs)
            if self.core and ret:
                self.core.refresh_window()
            return ret
        return wrap

    def always(self, func):
        """
        Decorator that refreshs the UI no matter what after the function
        """
        def wrap(*args, **kwargs):
            ret = func(*args, **kwargs)
            if self.core:
                self.core.refresh_window()
            return ret
        return wrap

    def update(self, funct):
        """
        Decorator that only updates the screen
        """
        def wrap(*args, **kwargs):
            ret = func(*args, **kwargs)
            if self.core:
                self.core.doupdate()
            return ret
        return wrap

refresh_wrapper = RefreshWrapper()



# Copyright 2010-2011 Florent Le Coz <louiz@louiz.org>
#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the zlib license. See the COPYING file.

"""
Implementation of the XEP-0045: Multi-User Chat.
Add some facilities that are not available on the XEP_0045
sleek plugin
"""

from xml.etree import cElementTree as ET

from common import safeJID
import logging
log = logging.getLogger(__name__)

NS_MUC_ADMIN = 'http://jabber.org/protocol/muc#admin'

def send_private_message(xmpp, jid, line):
    """
    Send a private message
    """
    jid = safeJID(jid)
    xmpp.send_message(mto=jid, mbody=line, mtype='chat')

def send_groupchat_message(xmpp, jid, line):
    """
    Send a message to the groupchat
    """
    jid = safeJID(jid)
    xmpp.send_message(mto=jid, mbody=line, mtype='groupchat')

def change_show(xmpp, jid, own_nick, show, status):
    """
    Change our 'Show'
    """
    jid = safeJID(jid)
    pres = xmpp.make_presence(pto='%s/%s' % (jid, own_nick))
    if show: # if show is None, don't put a <show /> tag. It means "available"
        pres['type'] = show
    if status:
        pres['status'] = status
    pres.send()

def change_subject(xmpp, jid, subject):
    """
    Change the room subject
    """
    jid = safeJID(jid)
    msg = xmpp.make_message(jid)
    msg['type'] = 'groupchat'
    msg['subject'] = subject
    msg.send()

def change_nick(core, jid, nick, status=None, show=None):
    """
    Change our own nick in a room
    """
    xmpp = core.xmpp
    presence = xmpp.make_presence(pshow=show, pstatus=status, pto=safeJID('%s/%s' % (jid, nick)))
    core.events.trigger('changing_nick', presence)
    presence.send()

def join_groupchat(core, jid, nick, passwd='', maxhistory=None, status=None, show=None, seconds=0):
    xmpp = core.xmpp
    stanza = xmpp.makePresence(pto='%s/%s' % (jid, nick), pstatus=status, pshow=show)
    x = ET.Element('{http://jabber.org/protocol/muc}x')
    if passwd:
        passelement = ET.Element('password')
        passelement.text = passwd
        x.append(passelement)
    if seconds:
        history = ET.Element('{http://jabber.org/protocol/muc}history')
        history.attrib['seconds'] = str(seconds)
        x.append(history)
    stanza.append(x)
    core.events.trigger('joining_muc', stanza)
    to = stanza["to"]
    stanza.send()
    xmpp.plugin['xep_0045'].rooms[jid] = {}
    xmpp.plugin['xep_0045'].ourNicks[jid] = to.resource

def leave_groupchat(xmpp, jid, own_nick, msg):
    """
    Leave the groupchat
    """
    jid = safeJID(jid)
    try:
        xmpp.plugin['xep_0045'].leaveMUC(jid, own_nick, msg)
    except KeyError:
        log.debug("muc.leave_groupchat: could not leave the room %s" % jid)

def set_user_role(xmpp, jid, nick, reason, role, callback=None):
    """
    (try to) Set the role of a MUC user
    (role = 'none': eject user)
    """
    jid = safeJID(jid)
    iq = xmpp.makeIqSet()
    query = ET.Element('{%s}query' % NS_MUC_ADMIN)
    item = ET.Element('{%s}item' % NS_MUC_ADMIN, {'nick':nick, 'role':role})
    if reason:
        reason_el = ET.Element('{%s}reason' % NS_MUC_ADMIN)
        reason_el.text = reason
        item.append(reason_el)
    query.append(item)
    iq.append(query)
    iq['to'] = jid
    if callback:
        return iq.send(block=False, callback=callback)
    try:
        return iq.send()
    except Exception as e:
        return e.iq

def set_user_affiliation(xmpp, muc_jid, affiliation, nick=None, jid=None, reason=None, callback=None):
    """
    (try to) Set the affiliation of a MUC user
    """
    muc_jid = safeJID(muc_jid)
    query = ET.Element('{http://jabber.org/protocol/muc#admin}query')
    if nick:
        item = ET.Element('{http://jabber.org/protocol/muc#admin}item', {'affiliation':affiliation, 'nick':nick})
    else:
        item = ET.Element('{http://jabber.org/protocol/muc#admin}item', {'affiliation':affiliation, 'jid':str(jid)})

    if reason:
        reason_item = ET.Element('{http://jabber.org/protocol/muc#admin}reason')
        reason_item.text = reason
        item.append(reason_item)

    query.append(item)
    iq = xmpp.makeIqSet(query)
    iq['to'] = muc_jid
    if callback:
        return iq.send(block=False, callback=callback)
    try:
        return xmpp.plugin['xep_0045'].setAffiliation(str(muc_jid), str(jid) if jid else None, nick, affiliation)
    except:
        import traceback
        log.debug('Error setting the affiliation: %s', traceback.format_exc())
        return False



# Copyright 2010-2011 Florent Le Coz <louiz@louiz.org>
#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the zlib license. See the COPYING file.

"""
Defines a Singleton function that initialize an object
of the given class if it was never instantiated yet. Else, returns
the previously instantiated object.
This method is the only one that I can come up with that do not call
__init__() each time.
"""

instances = {}
def Singleton(cls, *args, **kwargs):
    if not cls in instances:
        instances[cls] = cls(*args, **kwargs)
    return instances[cls]



from gettext import gettext as _

MOODS = {
          'afraid': _('Afraid'),
          'amazed': _('Amazed'),
          'angry': _('Angry'),
          'amorous': _('Amorous'),
          'annoyed': _('Annoyed'),
          'anxious': _('Anxious'),
          'aroused': _('Aroused'),
          'ashamed': _('Ashamed'),
          'bored': _('Bored'),
          'brave': _('Brave'),
          'calm': _('Calm'),
          'cautious': _('Cautious'),
          'cold': _('Cold'),
          'confident': _('Confident'),
          'confused': _('Confused'),
          'contemplative': _('Contemplative'),
          'contented': _('Contented'),
          'cranky': _('Cranky'),
          'crazy': _('Crazy'),
          'creative': _('Creative'),
          'curious': _('Curious'),
          'dejected': _('Dejected'),
          'depressed': _('Depressed'),
          'disappointed': _('Disappointed'),
          'disgusted': _('Disgusted'),
          'dismayed': _('Dismayed'),
          'distracted': _('Distracted'),
          'embarrassed': _('Embarrassed'),
          'envious': _('Envious'),
          'excited': _('Excited'),
          'flirtatious': _('Flirtatious'),
          'frustrated': _('Frustrated'),
          'grumpy': _('Grumpy'),
          'guilty': _('Guilty'),
          'happy': _('Happy'),
          'hopeful': _('Hopeful'),
          'hot': _('Hot'),
          'humbled': _('Humbled'),
          'humiliated': _('Humiliated'),
          'hungry': _('Hungry'),
          'hurt': _('Hurt'),
          'impressed': _('Impressed'),
          'in_awe': _('In awe'),
          'in_love': _('In love'),
          'indignant': _('Indignant'),
          'interested': _('Interested'),
          'intoxicated': _('Intoxicated'),
          'invincible': _('Invincible'),
          'jealous': _('Jealous'),
          'lonely': _('Lonely'),
          'lucky': _('Lucky'),
          'mean': _('Mean'),
          'moody': _('Moody'),
          'nervous': _('Nervous'),
          'neutral': _('Neutral'),
          'offended': _('Offended'),
          'outraged': _('Outraged'),
          'playful': _('Playful'),
          'proud': _('Proud'),
          'relaxed': _('Relaxed'),
          'relieved': _('Relieved'),
          'remorseful': _('Remorseful'),
          'restless': _('Restless'),
          'sad': _('Sad'),
          'sarcastic': _('Sarcastic'),
          'serious': _('Serious'),
          'shocked': _('Shocked'),
          'shy': _('Shy'),
          'sick': _('Sick'),
          'sleepy': _('Sleepy'),
          'spontaneous': _('Spontaneous'),
          'stressed': _('Stressed'),
          'strong': _('Strong'),
          'surprised': _('Surprised'),
          'thankful': _('Thankful'),
          'thirsty': _('Thirsty'),
          'tired': _('Tired'),
          'undefined': _('Undefined'),
          'weak': _('Weak'),
          'worried': _('Worried')
}




ACTIVITIES = {
        'doing_chores': {
            'category': _('Doing_chores'),

            'buying_groceries': _('Buying groceries'),
            'cleaning': _('Cleaning'),
            'cooking': _('Cooking'),
            'doing_maintenance': _('Doing maintenance'),
            'doing_the_dishes': _('Doing the dishes'),
            'doing_the_laundry': _('Doing the laundry'),
            'gardening': _('Gardening'),
            'running_an_errand': _('Running an errand'),
            'walking_the_dog': _('Walking the dog'),
            'other': _('Other'),
            },
        'drinking': {
            'category': _('Drinking'),

            'having_a_beer': _('Having a beer'),
            'having_coffee': _('Having coffee'),
            'having_tea': _('Having tea'),
            'other': _('Other'),
            },
        'eating': {
            'category':_('Eating'),

            'having_breakfast': _('Having breakfast'),
            'having_a_snack': _('Having a snack'),
            'having_dinner': _('Having dinner'),
            'having_lunch': _('Having lunch'),
            'other': _('Other'),
            },
        'exercising': {
            'category': _('Exercising'),

            'cycling': _('Cycling'),
            'dancing': _('Dancing'),
            'hiking': _('Hiking'),
            'jogging': _('Jogging'),
            'playing_sports': _('Playing sports'),
            'running': _('Running'),
            'skiing': _('Skiing'),
            'swimming': _('Swimming'),
            'working_out': _('Working out'),
            'other': _('Other'),
            },
        'grooming': {
            'category': _('Grooming'),

            'at_the_spa': _('At the spa'),
            'brushing_teeth': _('Brushing teeth'),
            'getting_a_haircut': _('Getting a haircut'),
            'shaving': _('Shaving'),
            'taking_a_bath': _('Taking a bath'),
            'taking_a_shower': _('Taking a shower'),
            'other': _('Other'),
            },
        'having_appointment': {
            'category': _('Having appointment'),

            'other': _('Other'),
            },
        'inactive': {
            'category': _('Inactive'),

            'day_off': _('Day_off'),
            'hanging_out': _('Hanging out'),
            'hiding': _('Hiding'),
            'on_vacation': _('On vacation'),
            'praying': _('Praying'),
            'scheduled_holiday': _('Scheduled holiday'),
            'sleeping': _('Sleeping'),
            'thinking': _('Thinking'),
            'other': _('Other'),
            },
        'relaxing': {
            'category': _('Relaxing'),

            'fishing': _('Fishing'),
            'gaming': _('Gaming'),
            'going_out': _('Going out'),
            'partying': _('Partying'),
            'reading': _('Reading'),
            'rehearsing': _('Rehearsing'),
            'shopping': _('Shopping'),
            'smoking': _('Smoking'),
            'socializing': _('Socializing'),
            'sunbathing': _('Sunbathing'),
            'watching_a_movie': _('Watching a movie'),
            'watching_tv': _('Watching tv'),
            'other': _('Other'),
            },
        'talking': {
            'category': _('Talking'),

            'in_real_life': _('In real life'),
            'on_the_phone': _('On the phone'),
            'on_video_phone': _('On video phone'),
            'other': _('Other'),
            },
        'traveling': {
            'category': _('Traveling'),

            'commuting': _('Commuting'),
            'driving': _('Driving'),
            'in_a_car': _('In a car'),
            'on_a_bus': _('On a bus'),
            'on_a_plane': _('On a plane'),
            'on_a_train': _('On a train'),
            'on_a_trip': _('On a trip'),
            'walking': _('Walking'),
            'cycling': _('Cycling'),
            'other': _('Other'),
            },
        'undefined': {
            'category': _('Undefined'),

            'other': _('Other'),
            },
        'working': {
            'category': _('Working'),

            'coding': _('Coding'),
            'in_a_meeting': _('In a meeting'),
            'writing': _('Writing'),
            'studying': _('Studying'),
            'other': _('Other'),
            }
        }











"""
Insert a space between each character, in messages that you send.

Installation
------------

You only have to load the plugin:

.. code-block:: none
        /load spaces

"""
from plugin import BasePlugin

class Plugin(BasePlugin):
    def init(self):
        self.api.add_event_handler('muc_say', self.add_spaces)
        self.api.add_event_handler('conversation_say', self.add_spaces)
        self.api.add_event_handler('private_say', self.add_spaces)

    def add_spaces(self, msg, tab):
        msg['body'] = " ".join(x for x in msg['body'])



"""
This plugin allows you to ping an entity.

Installation
------------
You only have to load the plugin.

.. code-block:: none

    /load ping

Command
-------

.. glossary::

    /ping
        **Usage (globally):** ``/ping <jid>``

        **Usage (in a MUC tab):** ``/ping <jid or nick>``

        **Usage (in a conversation tab):** ``/ping [jid]``

        Globally, you can do ``/ping jid@example.com`` to get a ping.

        In a MUC, you can either do it to a JID or a nick (``/ping nick`` or ``/ping
        jid@example.com``).

        In a private or a direct conversation, you can do ``/ping`` to ping
        the current interlocutor.
"""

from plugin import BasePlugin
from roster import roster
from common import safeJID
import common
import tabs
import time


class Plugin(BasePlugin):
    def init(self):
        self.core.xmpp.register_plugin('xep_0199')
        self.core.xmpp.plugin['xep_0115'].update_caps()
        self.api.add_command('ping', self.command_ping,
                usage='<jid>',
                help='Send an XMPP ping to jid (see XEP-0199).',
                short='Send a ping',
                completion=self.completion_ping)
        self.api.add_tab_command(tabs.MucTab, 'ping', self.command_muc_ping,
                usage='<jid|nick>',
                help='Send an XMPP ping to jid or nick (see XEP-0199).',
                short='Send a ping.',
                completion=self.completion_muc_ping)
        for _class in (tabs.PrivateTab, tabs.ConversationTab):
            self.api.add_tab_command(_class, 'ping', self.command_private_ping,
                    usage='[jid]',
                    help='Send an XMPP ping to the current interlocutor or the given JID.',
                    short='Send a ping',
                    completion=self.completion_ping)

    def command_ping(self, arg):
        if not arg:
            return self.core.command_help('ping')
        jid = safeJID(arg)
        start = time.time()
        def callback(iq):
            delay = time.time() - start
            if iq['type'] == 'error' and iq['error']['condition'] in ('remote-server-timeout', 'remote-server-not-found'):
                self.api.information('%s did not respond to ping' % jid, 'Info')
            else:
                self.api.information('%s responded to ping after %s s' % (jid, round(delay, 4)), 'Info')

        self.core.xmpp.plugin['xep_0199'].send_ping(jid=jid, callback=callback)

    def completion_muc_ping(self, the_input):
        users = [user.nick for user in self.api.current_tab().users]
        l = [contact.bare_jid for contact in roster.get_contacts()]
        users.extend(l)
        return the_input.auto_completion(users, '', quotify=False)

    def command_private_ping(self, arg):
        if arg:
            return self.command_ping(arg)
        self.command_ping(self.api.current_tab().get_name())

    def command_muc_ping(self, arg):
        if not arg.strip():
            return
        user = self.api.current_tab().get_user_by_name(arg)
        if user:
            jid = safeJID(self.api.current_tab().get_name())
            jid.resource = user.nick
        else:
            jid = safeJID(arg)
        self.command_ping(jid.full)

    def completion_ping(self, the_input):
        l = [contact.bare_jid for contact in roster.get_contacts()]
        return the_input.auto_completion(l, '', quotify=False)

